/* Based on:
 * $Id: sftp.c,v 1.17 2009/04/28 10:35:30 bagder Exp $
 * 
 * Sample showing how to do SFTP transfers.
 * 
 * The sample code has default values for host name, user name,
 * password and path to copy, but you can specify them
 * on the command line like:
 * 
 * sftpmrr -h:172.16.22.192 -u:user -p:password -f:localfile -r:remotefile
 * -a:{get|put} -port:tcpport -compression:0thru9"
 */

#include <string>
#include "libssh2_config.h"
#include <libssh2.h>
#include <libssh2_sftp.h>
#include <stdlib.h>

#ifdef HAVE_WINSOCK2_H
# include <winsock2.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif
# ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>

#ifdef WIN32
#include <io.h>
#endif

struct struct_settings {
   std::string     host;
   std::string     user;
   std::string     password;
   std::string     localfile;
   std::string     remotefile;
   std::string     action;
   int             bufsize;
   int             port;
   int             compression;
};

const char     *keyfile1 = "~/.ssh/id_rsa.pub";
const char     *keyfile2 = "~/.ssh/id_rsa";

#if 0
const char     *username;
const char     *settings.password.c_str() = "settings.password.c_str()";
   const char     *sftppath = "/tmp/TEST";
   const char     *localfile = "mydata.bin";

#endif

   static void     kbd_callback(const char *name, int name_len,
                  const char *instruction, int instruction_len, int num_prompts,
                  const LIBSSH2_USERAUTH_KBDINT_PROMPT * prompts,
                  LIBSSH2_USERAUTH_KBDINT_RESPONSE * responses,
                  void **abstract)
{
   int             i;
   size_t          n;
   char            buf[1024];

   (void) abstract;

   printf("Performing keyboard-interactive authentication.\n");

   printf("Authentication name: '");
   fwrite(name, 1, name_len, stdout);
   printf("'\n");

   printf("Authentication instruction: '");
   fwrite(instruction, 1, instruction_len, stdout);
   printf("'\n");

   printf("Number of prompts: %d\n\n", num_prompts);

   for (i = 0; i < num_prompts; i++) {
      printf("Prompt %d from server: '", i);
      fwrite(prompts[i].text, 1, prompts[i].length, stdout);
      printf("'\n");

      printf("Please type response: ");
      fgets(buf, sizeof(buf), stdin);
      n = strlen(buf);
      while (n > 0 && strchr("\r\n", buf[n - 1]))
         n--;
      buf[n] = 0;

      responses[i].text = strdup(buf);
      responses[i].length = n;

      printf("Response %d from user is '", i);
      fwrite(responses[i].text, 1, responses[i].length, stdout);
      printf("'\n\n");
   }

   printf("Done. Sending keyboard-interactive responses to server now.\n");
}

/*---  function MatchArgName  --------------------------
 *  Determine whether an argument matches a command.
 *
 *  Entry:  pszArgFromCmdline    is an argument like "a" or "cc:off".
 *
 *  Exit:   strArgName  holds the name of the argument, like "a" or "cc".
 *                              strArgVal       holds the value of the argument after the :, if any.
 *          We stop scanning the argument from the command line at the first
 *          ':' or NUL.
 */
void            CrackArg(const char *pszArgFromCmdline, std::string & strArgName, std::string & strArgVal)
{
   const char     *pszColon = strchr(pszArgFromCmdline, ':');

   if (pszColon) {
      strArgName = std::string(pszArgFromCmdline, pszColon - pszArgFromCmdline);
      strArgVal = std::string(1 + pszColon);
   } else {
      strArgName = pszArgFromCmdline;
      strArgVal = "";
   }
}

bool            ParseCmdline(int argc, char *argv[], struct_settings & settings)
{
   bool            bOK = true;

   for (int iarg = 1; iarg < argc; iarg++) {
      std::string     strArgName, strArgVal;
      char           *pszArg = argv[iarg];

      if ('-' == pszArg[0]) {
         CrackArg(1 + pszArg, strArgName, strArgVal);
         if ("h" == strArgName) {
            settings.host = strArgVal;
         } else if ("a" == strArgName) {
            settings.action = strArgVal;
         } else if ("u" == strArgName) {
            settings.user = strArgVal;
         } else if ("p" == strArgName) {
            settings.password = strArgVal;
         } else if ("b" == strArgName) {
            settings.bufsize = atol(strArgVal.c_str());
         } else if ("f" == strArgName) {
            settings.localfile = strArgVal;
         } else if ("r" == strArgName) {
            settings.remotefile = strArgVal;
         } else if ("port" == strArgName) {
            settings.port = atoi(strArgVal.c_str());
         } else if ("compression" == strArgName) {
            settings.compression = atoi(strArgVal.c_str());
         } else {
            printf("Unrecognized option: %s\n", strArgName.c_str());
            bOK = false;
         }
      } else {
         printf("Unrecognized argument: %s\n", strArgName.c_str());
         bOK = false;
      }
   }
   if (0 == settings.host.length() || 0 == settings.localfile.length()
     || 0 == settings.password.length() || 0 == settings.remotefile.length()
     || 0 == settings.user.length() || 0 == settings.action.length()) {
      bOK = false;
   }
   return bOK;
}

void            usage()
{
   puts("Usage: sftpmrr -h:host -u:user -p:password -f:localfile -r:remotepath ");
   puts("  -a:{get|put} -port:tcpport -compression:0thru9");
   puts("e.g. WeUpload -h:192.168.20.22 -u:fred -p:secret -f:c:\\tmp\\hi.txt");
   puts("  -r:/home/fred/hi.txt -b:32768 -port:22 -compression:6");
}

#ifndef WIN32
// Non-Windows version of GetTickCount.
// It returns milliseconds, but it's accurate only to the second.
unsigned long   GetTickCount()
{
   time_t          mytime = time(NULL);

   // To avoid integer overflow, we look only at the least
   // significant N bits of the timestamp before multiplying.
   unsigned long   ms = (mytime % 1000000) * 1000;

   return ms;
}

#define _S_IWRITE 0
#define _S_IREAD  0
#define O_BINARY  0
#endif

int             main(int argc, char *argv[])
{
   unsigned long   hostaddr;
   int             sock, i, auth_pw = 0, remotemode;
   struct sockaddr_in sin;
   const char     *fingerprint;
   char           *userauthlist;
   LIBSSH2_SESSION *session;
   int             rc;
   LIBSSH2_SFTP   *sftp_session;
   LIBSSH2_SFTP_HANDLE *sftp_handle;
   int             hand;
   unsigned long   dwmsStart, dwmsElapsed, filesize = 0;
   double          secs, mb_per_sec;

#ifdef WIN32
   WSADATA         wsadata;

   WSAStartup(MAKEWORD(2, 0), &wsadata);
#endif
   struct struct_settings settings;

   settings.bufsize = 1000000;
   settings.port = 22;
   settings.compression = 0;
   bool            bOK = ParseCmdline(argc, argv, settings);

   if (!bOK) {
      usage();
      exit(1);
   }
   rc = libssh2_init(0);
   if (rc != 0) {
      fprintf(stderr, "libssh2 initialization failed (%d)\n", rc);
      return 1;
   }
   /*
    * The application code is responsible for creating the socket and
    * establishing the connection
    */
   sock = socket(AF_INET, SOCK_STREAM, 0);

   hostaddr = inet_addr(settings.host.c_str());

   sin.sin_family = AF_INET;
   sin.sin_port = htons(22);
   sin.sin_addr.s_addr = hostaddr;
   if (connect(sock, (struct sockaddr *) (&sin),
       sizeof(struct sockaddr_in)) != 0) {
      fprintf(stderr, "failed to connect!\n");
      return -1;
   }
   /*
    * Create a session instance
    */

   session = libssh2_session_init();
   if (!session)
      return -1;

   if (settings.compression > 0) {
      libssh2_session_method_pref(session, LIBSSH2_METHOD_COMP_CS, "zlib");
      libssh2_session_method_pref(session, LIBSSH2_METHOD_COMP_SC, "zlib");
      printf("Enabling compression.\n");
   } else {
      printf("Not enabling compression.\n");
   }

   /* Since we have set non-blocking, tell libssh2 we are blocking */
   libssh2_session_set_blocking(session, 1);

   /*
    * ... start it up. This will trade welcome banners, exchange keys,
    * and setup crypto, compression, and MAC layers
    */
   rc = libssh2_session_startup(session, sock);
   if (rc) {
      fprintf(stderr, "Failure establishing SSH session: %d\n", rc);
      return -1;
   }
   /*
    * At this point we havn't yet authenticated.  The first thing to do
    * is check the hostkey's fingerprint against our known hosts Your
    * app may have it hard coded, may go to a file, may present it to
    * the user, that's your call
    */
   fingerprint = libssh2_hostkey_hash(session, LIBSSH2_HOSTKEY_HASH_SHA1);
   fprintf(stderr, "Fingerprint: ");
   for (i = 0; i < 20; i++) {
      fprintf(stderr, "%02X ", (unsigned char) fingerprint[i]);
   }
   fprintf(stderr, "\n");

   /* check what authentication methods are available */
   userauthlist = libssh2_userauth_list(session, settings.user.c_str(), strlen(settings.user.c_str()));
   printf("Authentication methods: %s\n", userauthlist);
   if (strstr(userauthlist, "password") != NULL) {
      auth_pw |= 1;
   }
   if (strstr(userauthlist, "keyboard-interactive") != NULL) {
      auth_pw |= 2;
   }
   if (strstr(userauthlist, "publickey") != NULL) {
      auth_pw |= 4;
   }
#if 0
   /* if we got an 4. argument we set this option if supported */
   if (argc > 5) {
      if ((auth_pw & 1) && !strcasecmp(argv[5], "-p")) {
         auth_pw = 1;
      }
      if ((auth_pw & 2) && !strcasecmp(argv[5], "-i")) {
         auth_pw = 2;
      }
      if ((auth_pw & 4) && !strcasecmp(argv[5], "-k")) {
         auth_pw = 4;
      }
   }
#endif

   if (auth_pw & 1) {
      /* We could authenticate via settings.password.c_str() */
      if (libssh2_userauth_password(session, settings.user.c_str(), settings.password.c_str())) {
         fprintf(stderr, "Authentication by settings.password.c_str() failed.\n");
         goto shutdown;
      }
   } else if (auth_pw & 2) {
      /* Or via keyboard-interactive */
      if (libssh2_userauth_keyboard_interactive(session, settings.user.c_str(), &kbd_callback)) {
         printf("\tAuthentication by keyboard-interactive failed!\n");
         goto shutdown;
      } else {
         printf("\tAuthentication by keyboard-interactive succeeded.\n");
      }
   } else if (auth_pw & 4) {
      /* Or by public key */
      if (libssh2_userauth_publickey_fromfile(session, settings.user.c_str(), keyfile1, keyfile2, settings.password.c_str())) {
         printf("\tAuthentication by public key failed!\n");
         goto shutdown;
      } else {
         printf("\tAuthentication by public key succeeded.\n");
      }
   } else {
      printf("No supported authentication methods found!\n");
      goto shutdown;
   }

   fprintf(stderr, "libssh2_sftp_init()!\n");
   sftp_session = libssh2_sftp_init(session);

   if (!sftp_session) {
      fprintf(stderr, "Unable to init SFTP session\n");
      goto shutdown;
   }
   fprintf(stderr, "libssh2_sftp_open()!\n");
   /* Request a file via SFTP */
   remotemode = LIBSSH2_FXF_READ;
   if ("put" == settings.action)
      remotemode = LIBSSH2_FXF_WRITE | LIBSSH2_FXF_CREAT;
   sftp_handle =
     libssh2_sftp_open(sftp_session, settings.remotefile.c_str(), remotemode, 0777);

   if (!sftp_handle) {
      fprintf(stderr, "Unable to open file with SFTP\n");
      goto shutdown;
   }
   fprintf(stderr, "libssh2_sftp_open() succeeded.\n");
   dwmsStart = GetTickCount();
   if ("get" == settings.action) {
      hand = open(settings.localfile.c_str(), O_BINARY | O_CREAT | O_RDWR, _S_IWRITE);
      // On Ubuntu to Solaris, get gets 9.87 MB/sec!
      do {
         char            mem[65536];

         /* loop until we fail */
         // fprintf(stderr, "libssh2_sftp_read()!\n");
         rc = libssh2_sftp_read(sftp_handle, mem, sizeof(mem));
         if (rc > 0) {
            write(hand, mem, rc);
            filesize += rc;
         } else {
            printf("libssh2_sftp_read with code %d\n", rc);
            break;
         }
      } while (1);
   } else if ("put" == settings.action) {
      hand = open(settings.localfile.c_str(), O_BINARY | O_RDONLY, _S_IREAD);
      if (-1 == hand) {
         perror("opening local file");
         exit(2);
      }
      int             nwrites = 0;

      // On Ubuntu to Solaris:
      // 1.85 MB/sec upload with 32500 bufsize.
      // 1.4 with 65536 (because it gets split into chunks,
      // the last of which is small).
      // 1.35 MB/sec with 16384.
      // 32500 seems to be the buffer size that actually gets used.
      char            mem[32500];

      printf("Using output buffer size of %d bytes\n", sizeof(mem));
      do {
         /* loop until we fail */
         int             nbytes = read(hand, mem, sizeof(mem));

         if (nbytes > 0) {
            filesize += nbytes;
            int             offset = 0, bytes_to_send = nbytes;

            do {
               rc = libssh2_sftp_write(sftp_handle, mem + offset, bytes_to_send);
               if (rc < 0) {
                  printf("libssh2_sftp_write failed with code %d\n", rc);
                  break;
               }
               ++nwrites;
               if (0 == (nwrites % 100)) {
                  printf("libssh2_sftp_write %d sent %d bytes\n", nwrites, rc);
               }
               offset += rc;
               bytes_to_send -= rc;
            } while (bytes_to_send > 0);
         } else {
            break;
         }
      } while (rc >= 0);
   } else {
      puts("action must be get or put.");
      exit(1);
   }
   close(hand);
   libssh2_sftp_close(sftp_handle);

   dwmsElapsed = GetTickCount() - dwmsStart;
   secs = 0.001 * dwmsElapsed;
   printf("Transferred %.3f MB\n", ((double) filesize) / (1024.0 * 1024.0));
   mb_per_sec = (filesize / secs) / (1024.0 * 1024.0);
   printf("%s took %.3f seconds = %.2f MB/sec\n", settings.action.c_str(), secs, mb_per_sec);
   libssh2_sftp_shutdown(sftp_session);

shutdown:

   libssh2_session_disconnect(session, "Normal Shutdown, Thank you for playing");
   libssh2_session_free(session);

#ifdef WIN32
   closesocket(sock);
#else
   close(sock);
#endif
   fprintf(stderr, "all done\n");

   libssh2_exit();

   return 0;
}
